import logging
from contextlib import contextmanager

import grpc
from fastapi import HTTPException


logger = logging.getLogger(__name__)


@contextmanager
def grpc_error_handling():
    try:
        yield
    except grpc.RpcError as rpc_error:
        # Handle gRPC errors here
        status_code = rpc_error.code()
        details = rpc_error.details()
        logger.warning(f"gRPC error: {status_code}, {details}")
        if (
            status_code == grpc.StatusCode.INVALID_ARGUMENT
            or status_code == grpc.StatusCode.UNIMPLEMENTED
        ):
            raise HTTPException(status_code=400, detail=details)

        # If the error is another type of gRPC error, we should return a 500
        if status_code == grpc.StatusCode.UNAVAILABLE:
            logger.warning("Server is unavailable, please try again later.")
        elif status_code == grpc.StatusCode.INTERNAL:
            logger.warning("Internal server error occurred.")
        else:
            logger.warning(f"An unexpected error occurred: {status_code}")
        raise HTTPException(status_code=500, detail="Briton error during inference")
    except Exception as ex:
        raise HTTPException(status_code=500, detail=f"An error has occurred: {ex}")



import os
from itertools import count
from pathlib import Path
from typing import Any, Dict, cast

import fastapi
from transformers import AutoTokenizer, PreTrainedTokenizerFast

from briton.briton import BritonInteractor, BritonInteractorImpl
from briton.checks import trllm_config_check
from briton.constants import (
    DEFAULT_BRITON_PORT,
    DEFAULT_MAX_FSM_WORKERS,
    FSM_CACHE_DIR,
    MODEL_INPUT_TO_BRITON_FIELD,
    OPENAI_COMPATIBLE_TAG,
    TOOL_CALL_IDS,
    TOOL_CALL_TOKENS,
)
from briton.error_handling import grpc_error_handling
from briton.fsm_cache import FsmCache, add_schema_to_cache
from briton.proto import InferenceRequest
from briton.schema import (
    create_completion,
    create_completion_chunks,
    get_prompt,
    update_raw_model_input,
    validate_model_input,
)
from briton.secrets import get_hf_token_or_none
from briton.trtllm_build_config import TrussTRTLLMBuildConfiguration


# TODO: Decide configurable interface and remove this
BRITON_SEND_OPENAI_RESPONSE = False


class Model:
    def __init__(self, **kwargs):
        self._loaded = False
        self._model = None
        self._config = kwargs["config"]
        self._data_dir = kwargs["data_dir"]
        self._stub = None
        self._secrets = kwargs["secrets"]
        self._request_id_counter = count(start=1)
        self._briton_process = None
        model_metadata = self._config.get("model_metadata", {})
        # TODO: 1) This is not checking model_metadata["tags"], where the tag should be.
        #       2) What does the tag do/mean? We aren't using it for anything in this file.
        #       3) Decide how to actually handle switching between openai and non-openai responses.
        if OPENAI_COMPATIBLE_TAG not in self._config.get("model_metadata", {}):
            print(
                "Warning: Defaulting to openai compatible interface, even though not indicated in config."
            )

        trllm_config_check(self._config)
        trtllm_config = self._config.get("trt_llm")
        truss_trtllm_build_config = TrussTRTLLMBuildConfiguration(**trtllm_config.get("build"))
        self._base_model = truss_trtllm_build_config.base_model
        self._tp_count = truss_trtllm_build_config.tensor_parallel_count
        self._tokenizer_repository = truss_trtllm_build_config.checkpoint_repository.repo
        self._kv_cache_free_gpu_mem_fraction = (
            truss_trtllm_build_config.kv_cache_free_gpu_mem_fraction
        )
        self._enable_kv_cache_reuse = (
            truss_trtllm_build_config.plugin_configuration.use_paged_context_fmha
        )
        self._enable_chunked_context = truss_trtllm_build_config.enable_chunked_context

        self._max_num_tokens = truss_trtllm_build_config.max_num_tokens

        self._hf_token = get_hf_token_or_none(self._secrets)

        # TODO(@bdubayah): configure this based on CPU. But os.cpu_count() returns the
        # number of CPUs for the entire node, not just the container.
        self._max_fsm_workers = DEFAULT_MAX_FSM_WORKERS
        print(f"Using {self._max_fsm_workers} workers for FSM schema generation")
        # Allow passing briton_interactor for ease of testing
        self._briton_interactor: BritonInteractor = model_metadata.get(
            "briton_interactor", BritonInteractorImpl()
        )
        self._tool_call_token = TOOL_CALL_TOKENS.get(self._base_model)
        self._tool_call_token_id = TOOL_CALL_IDS.get(self._base_model)

    def load(self):
        if self._loaded:
            return

        # TODO(pankaj) Support loading bundled tokenizer rather than from HF
        self._tokenizer = cast(
            PreTrainedTokenizerFast,
            self._briton_interactor.auto_tokenizer_from_pretrained(
                self._tokenizer_repository, hf_token=self._hf_token
            ),
        )

        # We only support Llama and mistral with Briton, for which this should
        # apply.
        assert isinstance(self._tokenizer, PreTrainedTokenizerFast)

        # These are tokens outside of tokenizer.json. We need to pass these to
        # Briton, to pass to rust tokenizer.
        added_token_decoders = self._tokenizer.added_tokens_decoder
        added_tokens = [token for token in added_token_decoders.values()]

        self._fsm_cache = FsmCache(
            self._briton_interactor.fsm_cache_dir(),
            self._tokenizer,
            self._max_fsm_workers,
            self._tool_call_token_id,
        )

        load_briton = self._briton_interactor.load
        load_briton(
            model_name="briton",
            engine_path=self._data_dir,
            hf_tokenizer=self._tokenizer_repository,
            work_dir=self._data_dir,
            fsm_cache_dir=self._fsm_cache.cache_dir,
            kv_cache_free_gpu_mem_fraction=self._kv_cache_free_gpu_mem_fraction,
            port=DEFAULT_BRITON_PORT,
            added_tokens=added_tokens,
            max_num_tokens=self._max_num_tokens,
            enable_chunked_context=self._enable_chunked_context,
            hf_token=self._hf_token,
            tp_count=self._tp_count,
        )
        self._loaded = True

    async def predict(self, model_input: Dict[str, Any], request: fastapi.Request):
        # TODO(pankaj) Wire up request cancellation
        """
        Run inference

        Note that the async nature of this function is a little tricky. Care is
        needed to make sure this function is a regular async function and not an
        async generator, i.e. there shouldn't be any direct yields in this
        function. This is because we need to support both streaming and
        non-streaming cases in this function. We do this by either returning an
        async-generator for the streaming case, or directly the full text for
        the other case. Returning an async generator for non-streaming case
        interferes with the open ai client proxy.
        """
        if self._stub is None:
            self._stub = self._briton_interactor.create_grpc_stub(DEFAULT_BRITON_PORT)

        validated_input = validate_model_input(model_input)
        prompt = get_prompt(validated_input, self._tokenizer)
        model_input.pop("messages", None)

        request_id = self._calc_request_id()
        briton_request = InferenceRequest(
            request_id=request_id,
            input_text=prompt,
        )
        self._update_request_end_id_pad_id(briton_request, model_input)

        schema = validated_input.output_json_schema
        if schema is not None:
            briton_request.output_schema_hash = await add_schema_to_cache(self._fsm_cache, schema)
            force_tools = validated_input.force_tools
            if force_tools is not None:
                briton_request.force_tools = force_tools

        update_raw_model_input(model_input, validated_input)

        set_briton_request_fields_from_model_input(model_input, briton_request)
        for words in ["bad_words", "stop_words"]:
            if words in model_input:
                for word in model_input[words].split(","):
                    getattr(briton_request, words).append(word)

        resp_iter = self._stub.Infer(briton_request)

        # Get model name for openai compatibility
        model_name = validated_input.model if validated_input.model else ""

        with grpc_error_handling():
            if model_input.get("stream", True):
                # Advance the iterator to get the first chunk, to allow any validation error to be thrown
                async for first_chunk in resp_iter:
                    break

                async def generate_after_first_chunk():
                    yield first_chunk.output_text
                    async for chunk in resp_iter:
                        yield chunk.output_text

                async def generate_processed_after_first_chunk():
                    # Try to remove tool call token from first chunk only
                    output_text = first_chunk.output_text
                    if self._tool_call_token is not None:
                        output_text = output_text.removeprefix(self._tool_call_token)
                    yield output_text

                    eos_token = self._eos_token()
                    async for chunk in resp_iter:
                        output_text = chunk.output_text
                        if eos_token is not None:
                            output_text = output_text.removesuffix(eos_token)
                        yield output_text

                if BRITON_SEND_OPENAI_RESPONSE:
                    # TODO: Copy token count logic from spec dec truss
                    return create_completion_chunks(
                        req_id=str(request_id),
                        model=model_name,
                        input_text=generate_after_first_chunk(),
                        eos_token=self._eos_token(),
                        tool_token=self._tool_call_token,
                    )
                else:
                    return generate_processed_after_first_chunk()
            else:
                full_text = await _collect_text(resp_iter)
                if BRITON_SEND_OPENAI_RESPONSE:
                    # TODO: Copy token count logic from spec dec truss
                    return create_completion(
                        req_id=str(request_id),
                        model=model_name,
                        input_text=full_text,
                        eos_token=self._eos_token(),
                        tool_token=self._tool_call_token,
                    )
                else:
                    eos_token = self._eos_token()
                    if eos_token is not None:
                        full_text = full_text.removesuffix(eos_token)
                    if self._tool_call_token is not None:
                        full_text = full_text.removeprefix(self._tool_call_token)
                    return full_text

    def _calc_request_id(self) -> int:
        """Calculate unique request id.

        Not thread safe, but safe to use in single threaded async context. There
        are no async operations here, so this function is unlikely to be
        preempted in the middle. This is important otherwise we may end up with
        duplicate ids.
        """
        return int(str(os.getpid()) + str(next(self._request_id_counter)))

    def _eos_token(self):
        return getattr(self._tokenizer, "eos_token", None)

    def _eos_token_id(self):
        return getattr(self._tokenizer, "eos_token_id", None)

    def _pad_token_id(self):
        return getattr(self._tokenizer, "pad_token_id", None)

    def _update_request_end_id_pad_id(self, request, model_input):
        end_id = model_input.get("end_id", None) or self._eos_token_id()
        if end_id is not None:
            request.end_id = end_id
        pad_id = model_input.get("pad_id", None) or self._pad_token_id()
        if pad_id is not None:
            request.pad_id = pad_id


def set_briton_request_fields_from_model_input(model_input, briton_request):
    for model_input_key, briton_field in MODEL_INPUT_TO_BRITON_FIELD.items():
        if model_input_key in model_input:
            model_input_value = model_input[model_input_key]
            setattr(briton_request, briton_field, model_input_value)


async def _collect_text(async_text_iter):
    full_text = ""
    async for delta in async_text_iter:
        full_text += delta.output_text
    return full_text



import asyncio
import logging
import os
import signal
import threading
import time
from itertools import count
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

import fastapi
from fastapi import HTTPException as FastApiHTTPException
from transformers import AutoTokenizer, PreTrainedTokenizerFast

from briton.briton import (
    BritonInteractor,
    BritonInteractorImpl,
    EosRespondingBritonInteractor,
    FailingBritonInteractor,
)
from briton.constants import (
    DEFAULT_BRITON_PORT,
    DEFAULT_MAX_FSM_WORKERS,
    DEFAULT_TP_COUNT,
    FSM_CACHE_DIR,
    MODEL_INPUT_TO_BRITON_FIELD,
    OPENAI_COMPATIBLE_TAG,
    TOOL_CALL_IDS,
    TOOL_CALL_TOKENS,
)
from briton.data_structures import common_prefix_length
from briton.error_handling import grpc_error_handling
from briton.flags import TRUSS_DEVELOPMENT_MODE
from briton.fsm_cache import FsmCache, add_schema_to_cache
from briton.proc_utils import kill_current_process_and_children
from briton.proto import Batch, InferenceRequest
from briton.schema import (
    create_completion,
    create_completion_chunks,
    get_prompt,
    update_raw_model_input,
    validate_model_input,
)
from briton.secrets import get_hf_token_or_none
from briton.spec_dec import (
    SpecDecRequest,
    handle_request_exception,
    monitor_thread_fn,
    pick_tokens_from_end,
    test_predict,
    worker,
)


TARGET_PORT = DEFAULT_BRITON_PORT
DRAFT_PORT = TARGET_PORT + 1

# Relative to data dir
TARGET_MODEL_PATH = "target"
DRAFT_MODEL_PATH = "draft"

logger = logging.getLogger(__name__)


class Model:
    def __init__(self, **kwargs):
        self._loaded = False
        self._model = None
        self._config = kwargs["config"]
        self._data_dir = kwargs["data_dir"]
        self._secrets = kwargs["secrets"]
        self._request_id_counter = count(start=1)
        model_metadata = self._config.get("model_metadata", {})
        _openai_compatiblilty_check(model_metadata)

        self._tp_count = model_metadata.get("tp_count", DEFAULT_TP_COUNT)
        self._hf_tokenizer = _get_model_metadata_or_raise(model_metadata, "hf_tokenizer")
        self._base_model = _get_model_metadata_or_raise(model_metadata, "base_model")
        self._tool_call_token = TOOL_CALL_TOKENS.get(self._base_model)
        self._tool_call_token_id = TOOL_CALL_IDS.get(self._base_model)
        if not TRUSS_DEVELOPMENT_MODE:
            self._hf_spec_dec_repo_id = _get_model_metadata_or_raise(
                model_metadata, "hf_spec_dec_repo_id"
            )

        self._target_kv_cache_free_gpu_mem_fraction = _get_model_metadata_or_raise(
            model_metadata, "target_kv_cache_free_gpu_mem_fraction"
        )
        self._draft_kv_cache_free_gpu_mem_fraction = _get_model_metadata_or_raise(
            model_metadata, "draft_kv_cache_free_gpu_mem_fraction"
        )
        self._num_draft_max_input_tokens = _get_model_metadata_or_raise(
            model_metadata, "num_draft_max_input_tokens"
        )
        self._draft_max_num_tokens = model_metadata.get("draft_max_num_tokens", None)
        self._target_max_num_tokens = model_metadata.get("target_max_num_tokens", None)
        self._enable_chunked_context = model_metadata.get("enable_chunked_context", False)
        self._total_token_limit = model_metadata.get("total_token_limit", None)
        self._num_draft_tokens = _get_model_metadata_or_raise(model_metadata, "num_draft_tokens")
        self._hf_token = get_hf_token_or_none(self._secrets)
        self._lazy_init_done = False
        self._lazy_init_lock = None
        self._draft_stub = None
        self._target_stub = None
        self._draft_queue = None
        self._target_queue = None
        self._max_fsm_workers = DEFAULT_MAX_FSM_WORKERS

        # Total tokens input tokens in flight.
        # TODO(pankaj) Take generated tokens into account as well. That has to
        # be done carefully to avoid any performance impact. Going with this
        # simple mechanism for now.
        self._total_tokens_lock = None
        self._total_tokens = 0

        # Allow passing briton_interactor for ease of testing
        self._briton_interactor: BritonInteractor = model_metadata.get(
            "briton_interactor", BritonInteractorImpl()
        )

        # Briton interactor can also be passed as a string to pick a pre-built one
        if isinstance(self._briton_interactor, str):
            if self._briton_interactor == "EosRespondingBritonInteractor":
                logger.info("Using EosRespondingBritonInteractor")
                self._briton_interactor = EosRespondingBritonInteractor()
            elif self._briton_interactor == "FailingBritonInteractor":
                logger.info("Using FaiingBritonInteractor")
                self._briton_interactor = FailingBritonInteractor()
            else:
                raise ValueError(f"Unknown briton_interactor: {self._briton_interactor}")

    def load(self):
        if self._loaded:
            return

        # todo: Use special briton interactor for development mode
        self._briton_interactor.hf_download(
            self._hf_spec_dec_repo_id, local_dir=self._data_dir, hf_token=self._hf_token
        )

        # TODO(pankaj) Support loading bundled tokenizer rather than from HF
        self._tokenizer = cast(
            PreTrainedTokenizerFast,
            self._briton_interactor.auto_tokenizer_from_pretrained(
                self._hf_tokenizer, self._hf_token
            ),
        )

        # We only support Llama and mistral with Briton, for which this should
        # apply.
        assert isinstance(self._tokenizer, PreTrainedTokenizerFast)

        # These are tokens outside of tokenizer.json. We need to pass these to
        # Briton, to pass to rust tokenizer.
        added_token_decoders = self._tokenizer.added_tokens_decoder
        added_tokens = [token for token in added_token_decoders.values()]

        # Initialize n processes for _max_fsm_workers
        self._fsm_cache = FsmCache(
            self._briton_interactor.fsm_cache_dir(),
            self._tokenizer,
            self._max_fsm_workers,
            self._tool_call_token_id,
        )

        load_briton = self._briton_interactor.load
        load_briton(
            model_name="target",
            engine_path=(self._data_dir / TARGET_MODEL_PATH),
            hf_tokenizer=self._hf_tokenizer,
            work_dir=(self._data_dir / "target"),
            fsm_cache_dir=self._fsm_cache.cache_dir,
            kv_cache_free_gpu_mem_fraction=self._target_kv_cache_free_gpu_mem_fraction,
            port=TARGET_PORT,
            added_tokens=added_tokens,
            max_num_tokens=self._target_max_num_tokens,
            enable_chunked_context=self._enable_chunked_context,
            hf_token=self._hf_token,
            tp_count=self._tp_count,
        )
        load_briton(
            model_name="draft",
            engine_path=(self._data_dir / DRAFT_MODEL_PATH),
            hf_tokenizer=self._hf_tokenizer,
            work_dir=(self._data_dir / "draft"),
            fsm_cache_dir=self._fsm_cache.cache_dir,
            kv_cache_free_gpu_mem_fraction=self._target_kv_cache_free_gpu_mem_fraction,
            port=DRAFT_PORT,
            added_tokens=added_tokens,
            max_num_tokens=self._draft_max_num_tokens,
            enable_chunked_context=self._enable_chunked_context,
            hf_token=self._hf_token,
            tp_count=self._tp_count,
        )
        self._start_monitor()
        self._loaded = True

    def _start_monitor(self):
        # Overall, spec-dec model monitoring.
        monitor_settings = self._briton_interactor.monitor_settings()
        if not monitor_settings.start_monitor_thread:
            return

        def on_fail():
            logger.error("Spec dec model is stuck, exiting truss server")
            kill_current_process_and_children()

        async def monitor_fn():
            try:
                return await test_predict(self.predict)
            except Exception as e:
                return False

        self._monitor_thread = threading.Thread(
            target=monitor_thread_fn,
            args=(
                monitor_fn,
                on_fail,
                monitor_settings.should_continue,
                monitor_settings.period_secs,
                monitor_settings.max_retries,
                monitor_settings.timeout_secs,
                monitor_settings.retry_delay_secs,
            ),
            daemon=True,
        )
        self._monitor_thread.start()

    async def predict(self, model_input: Dict[str, Any], request: fastapi.Request):
        """
        Run inference

        Note that the async nature of this function is a little tricky. Care is
        needed to make sure this function is a regular async function and not an
        async generator, i.e. there shouldn't be any direct yields in this
        function. This is because we need to support both streaming and
        non-streaming cases in this function. We do this by either returning an
        async-generator for the streaming case, or directly the full text for
        the other case. Returning an async generator for non-streaming case
        interferes with the open ai client proxy.
        """

        async def is_cancelled_fn():
            disconnected = await request.is_disconnected()
            if disconnected:
                logger.info("Request disconnected, cancelling.")
            return disconnected

        if not self._lazy_init_done:
            # While this isn't completely safe, the async lock needs to be
            # created within the same async loop where it will be used. Ideally,
            # the proper solution would involve supporting asynchronous load
            # function, but that is not currently supported in Truss. The risk is
            # that multiple initial requests could end up with different lock
            # instances, making the lock ineffective. In practice, this is
            # highly unlikely. This issue could occur if one request executes
            # the line below and then gets preempted, allowing another request
            # to execute the same line. However, since there is no async
            # operation in the following line, it is very unlikely for the
            # request to be preempted at that point.
            if self._lazy_init_lock is None:
                self._lazy_init_lock = asyncio.Lock()

            async with self._lazy_init_lock:
                self._total_tokens_lock = asyncio.Lock()
                self._target_stub = self._briton_interactor.create_grpc_stub(TARGET_PORT)
                self._draft_stub = self._briton_interactor.create_grpc_stub(DRAFT_PORT)
                self._target_queue = asyncio.Queue()
                self._draft_queue = asyncio.Queue()
                asyncio.create_task(
                    worker(
                        self._target_queue,
                        self._target_consumer,
                        self._draft_queue,
                        self._draft_consumer,
                        asyncio.Event(),
                    )
                )
                self._lazy_init_done = True

        validated_input = validate_model_input(model_input)
        prompt = get_prompt(validated_input, self._tokenizer)
        model_input.pop("messages", None)
        input_ids = self._tokenizer(prompt).input_ids

        request_id = self._calc_request_id()
        spec_dec_request = SpecDecRequest(
            id=request_id,
            input_ids=input_ids,
            draft_ids=[],
            model_input=model_input,
            is_cancelled_fn=is_cancelled_fn,
        )

        schema = validated_input.output_json_schema
        if schema is not None:
            spec_dec_request.output_schema_hash = await add_schema_to_cache(self._fsm_cache, schema)
            force_tools = validated_input.force_tools
            if force_tools is not None:
                spec_dec_request.force_tools = force_tools

        update_raw_model_input(model_input, validated_input)

        # Wait for space in the total token limit.
        while self._total_tokens + len(input_ids) > self._total_token_limit:
            if await spec_dec_request.is_cancelled_fn():
                raise RuntimeError("Request cancelled")
            await asyncio.sleep(0.01)

        # Reserver space in total tokens before queueing
        async with self._total_tokens_lock:
            self._total_tokens += len(spec_dec_request.input_ids)
        self._target_queue.put_nowait(spec_dec_request)

        async def read_req_queue():
            while True:
                inc_text = await spec_dec_request.queue.get()
                if inc_text is None:
                    break
                if isinstance(inc_text, FastApiHTTPException):
                    raise inc_text
                if isinstance(inc_text, Exception):
                    raise inc_text
                yield inc_text

            # When done, release space from total tokens.
            async with self._total_tokens_lock:
                self._total_tokens -= len(spec_dec_request.input_ids)

        # Get model name for openai compatibility
        model_name = validated_input.model if validated_input.model else ""

        logger.debug(f"Input ids: {input_ids}")
        if validated_input.stream:
            return create_completion_chunks(
                req_id=str(request_id),
                model=model_name,
                input_text=read_req_queue(),
                eos_token=self._eos_token(),
                tool_token=self._tool_call_token,
                prompt_tokens=(
                    spec_dec_request.prompt_tokens if validated_input.include_stream_usage else None
                ),
                completion_tokens_fn=(
                    spec_dec_request.completion_tokens
                    if validated_input.include_stream_usage
                    else None
                ),
            )
        else:
            full_text = await _collect_text(read_req_queue())
            return create_completion(
                req_id=str(request_id),
                model=model_name,
                input_text=full_text,
                eos_token=self._eos_token(),
                tool_token=self._tool_call_token,
                prompt_tokens=spec_dec_request.prompt_tokens,
                completion_tokens=spec_dec_request.completion_tokens(),
            )

    def _calc_request_id(self) -> int:
        """Calculate unique request id.

        Not thread safe, but safe to use in single threaded async context. There
        are no async operations here, so this function is unlikely to be
        preempted in the middle. This is important otherwise we may end up with
        duplicate ids.
        """
        return int(str(os.getpid()) + str(next(self._request_id_counter)))

    async def _call_draft(
        self, request_id: int, spec_dec_request: SpecDecRequest, batch_request_ids: List[int]
    ):
        """Call draft model.

        This request is part of a batch of requests. All batch requests ids need to be
        provided in batch_request_ids.
        """
        # Draft model may only support a smaller context, so limit input tokens
        # to it.
        draft_tokens_start = pick_tokens_from_end(
            len(spec_dec_request.input_ids), self._num_draft_max_input_tokens
        )
        briton_request = InferenceRequest(
            request_id=request_id,
            input_ids=spec_dec_request.input_ids[draft_tokens_start:],
            request_output_len=self._num_draft_tokens,
            batch=Batch(request_ids=batch_request_ids),
            output_schema_hash=spec_dec_request.output_schema_hash,
            fsm_state=spec_dec_request.fsm_state,
            force_tools=spec_dec_request.force_tools,
        )
        self._update_request_end_id_pad_id(briton_request, spec_dec_request.model_input)
        with grpc_error_handling():
            resp_iter = self._draft_stub.Infer(briton_request)
            draft_output_ids = []
            async for delta in resp_iter:
                draft_output_ids.extend(delta.output_ids)
        return draft_output_ids

    async def _call_target(
        self,
        request_id: int,
        spec_dec_request: SpecDecRequest,
        batch_request_ids: List[int],
        max_tokens: Optional[int],
    ):
        """Call target model.

        This request is part of a batch of requests. All batch requests ids need to be
        provided in batch_request_ids.
        Proposed draft ids should be provided in draft_ids.
        Target model request needs all the sampling params.

        max_tokens is the maximum number of tokens to generate. Note that target model
        can generate more than one token due to draft token acceptance. max_tokens can
        be used to cut this short to adhere to the max tokens limit at request level.
        """
        request_output_len = self._num_draft_tokens + 1
        if max_tokens is not None:
            request_output_len = min(request_output_len, max_tokens)

        briton_request = InferenceRequest(
            request_id=request_id,
            input_ids=spec_dec_request.input_ids,
            draft_input_ids=spec_dec_request.draft_ids,
            request_output_len=request_output_len,
            batch=Batch(request_ids=batch_request_ids),
            output_schema_hash=spec_dec_request.output_schema_hash,
            fsm_state=spec_dec_request.fsm_state,
            force_tools=spec_dec_request.force_tools,
            write_output_states=True,  # The target should always write its state
        )
        model_input = spec_dec_request.model_input
        self._update_request_end_id_pad_id(briton_request, model_input)
        set_briton_request_fields_from_model_input(model_input, briton_request)
        for words in ["bad_words", "stop_words"]:
            if words in model_input:
                for word in model_input[words].split(","):
                    getattr(briton_request, words).append(word)
        with grpc_error_handling():
            resp_iter = self._target_stub.Infer(briton_request)
            target_text = ""
            target_output_ids = []
            target_output_states = []

            eos_encountered = False
            max_tokens_reached = False
            async for delta in resp_iter:
                # Include the eos token, but remove anything after it.
                if self._eos_token_id() in delta.output_ids:
                    eos_token_id_index = list(delta.output_ids).index(self._eos_token_id())
                    target_output_ids.extend(delta.output_ids[: (eos_token_id_index + 1)])
                    eos_index = delta.output_text.find(self._eos_token())
                    if eos_index != -1:
                        target_text += delta.output_text[: (eos_index + len(self._eos_token()))]
                    eos_encountered = True
                    break

                target_output_ids.extend(delta.output_ids)
                target_output_states.extend(delta.output_states)

                # Note that output_text may be empty string and that's ok.
                # Briton may be buffering for unicode characters.
                target_text += delta.output_text

                # TODO(pankaj): Cut exactly at max_tokens. We let the tokens go
                # a little bit over right now as mapping subset of tokens to
                # text is a little tricky.
                # Also ideally batch manager should honor the request_output_len above
                # but it doesn't seem to in this draft acceptance case.
                if max_tokens is not None and len(target_output_ids) >= max_tokens:
                    max_tokens_reached = True
                    break

        return (
            target_output_ids,
            target_output_states,
            target_text,
            eos_encountered,
            max_tokens_reached,
        )

    def _eos_token(self):
        return getattr(self._tokenizer, "eos_token", None)

    def _eos_token_id(self):
        return getattr(self._tokenizer, "eos_token_id", None)

    def _pad_token_id(self):
        return getattr(self._tokenizer, "pad_token_id", None)

    def _update_request_end_id_pad_id(
        self, briton_request: InferenceRequest, model_input: Dict[str, Any]
    ):
        end_id = model_input.get("end_id", None) or self._eos_token_id()
        if end_id is not None:
            briton_request.end_id = end_id
        pad_id = model_input.get("pad_id", None) or self._pad_token_id()
        if pad_id is not None:
            briton_request.pad_id = pad_id

    async def _target_consumer(self, requests: List[SpecDecRequest]) -> List[SpecDecRequest]:
        """Consumes requests from the target queue and fetches predictions from target model.

        Args:
            requests (List[Request]): Requests to process

        Returns:
            List[Request]: Requests to continue.
        """

        @handle_request_exception
        async def call_target(req, req_id, batch_request_ids):
            max_tokens = req.model_input.get("max_tokens", None)
            max_incremental_target_tokens = (
                max_tokens - req.produced_token_count if max_tokens is not None else None
            )
            st = time.time()
            (
                target_output_ids,
                target_output_states,
                target_text,
                eos_encountered,
                max_tokens_reached,
            ) = await self._call_target(
                req_id,
                req,
                batch_request_ids,
                max_incremental_target_tokens,
            )
            req.target_time_ms += (time.time() - st) * 1000
            # Streaming return target_text
            req.queue.put_nowait(target_text)
            req.produced_token_count += len(target_output_ids)
            req.iterations += 1
            accepted_count = common_prefix_length(req.draft_ids, target_output_ids)
            req.accepted_token_count += accepted_count

            def should_stop():
                if eos_encountered:
                    logger.debug("eos encountered")
                    return True

                if max_tokens_reached:
                    logger.debug("Reached max tokens, stopping")
                    return True

                if len(target_output_ids) == 0:
                    logger.debug("Target output is empty, stopping")
                    return True

                if target_output_ids == req.draft_ids:
                    logger.debug("Draft and target outputs are the same, stopping")
                    return True
                return False

            # Stopping conditions and cancellation handling.
            if should_stop() or await req.is_cancelled_fn():
                logger.info(f"Total generated token count: {req.produced_token_count}")
                logger.info(f"Total draft token count: {req.draft_token_count}")
                logger.info(f"Total draft time: {req.draft_time_ms}")
                logger.info(f"Total target time: {req.target_time_ms}")
                total_time = time.time() - req.start_time
                logger.info(f"Total time: {total_time}")
                logger.info(f"TPS: {req.produced_token_count / total_time}")
                logger.info(
                    f"Avg accepted tokens count: {req.accepted_token_count / req.iterations}"
                )
                req.queue.put_nowait(None)
                return None

            req.input_ids = req.input_ids + target_output_ids
            req.fsm_state = target_output_states[-1] if len(target_output_states) > 0 else None
            return req

        # Generate all request_ids upfront
        target_request_ids = {req.id: self._calc_request_id() for req in requests}
        batch_request_ids = target_request_ids.values()
        st = time.time()
        tasks = [
            asyncio.create_task(call_target(req, target_request_ids[req.id], batch_request_ids))
            for req in requests
        ]
        cont_reqs = await asyncio.gather(*tasks)
        logger.debug(f"all target reqs time", (time.time() - st) * 1000)
        cont_reqs = [req for req in cont_reqs if req is not None]
        return cont_reqs

    async def _draft_consumer(self, requests: List[SpecDecRequest]):
        """Consumes requests from the draft queue and fetches predictions from draft model.

        Args:
            requests (List[Request]): Requests to process

        Returns:
            List[Request]: Requests to continue.
        """

        @handle_request_exception
        async def call_draft(req, req_id, batch_request_ids):
            st = time.time()
            draft_ids = await self._call_draft(req_id, req, batch_request_ids)
            req.draft_time_ms += (time.time() - st) * 1000
            req.draft_ids = draft_ids
            req.draft_token_count += len(draft_ids)

        draft_request_ids = {req.id: self._calc_request_id() for req in requests}
        batch_request_ids = draft_request_ids.values()
        st = time.time()
        tasks = [
            asyncio.create_task(call_draft(req, draft_request_ids[req.id], batch_request_ids))
            for req in requests
        ]
        await asyncio.gather(*tasks)
        logger.debug(f"draft time for all reqs: ", (time.time() - st) * 1000)
        return requests


def set_briton_request_fields_from_model_input(
    model_input: Dict[str, Any], briton_request: InferenceRequest
):
    for model_input_key, briton_field in MODEL_INPUT_TO_BRITON_FIELD.items():
        if model_input_key in model_input:
            model_input_value = model_input[model_input_key]
            setattr(briton_request, briton_field, model_input_value)


async def _collect_text(async_text_iter):
    full_text = ""
    async for delta in async_text_iter:
        full_text += delta
    return full_text


def _openai_compatiblilty_check(model_metadata):
    if OPENAI_COMPATIBLE_TAG not in model_metadata:
        print(
            (
                "Warning: Defaulting to openai compatible interface, "
                "even though not indicated in config."
            )
        )
    return True


def _get_model_metadata_or_raise(model_metadata, key):
    if key not in model_metadata:
        raise ValueError(f"{key} is required in model_metadata")
    return model_metadata[key]



# TODO(pankaj) Start using this.
# Currently, this is not used, we use fastapi.HTTPException.
class HTTPException(Exception):
    def __init__(self, status_code, detail="HTTP Exception occurred"):
        self.status_code = status_code
        self.detail = detail
        super().__init__(f"{detail}: {status_code}")



import os
import signal

import psutil


def kill_current_process_and_children():
    current_process = psutil.Process(os.getpid())

    for child in current_process.children(recursive=True):
        os.kill(child.pid, signal.SIGKILL)

    os.kill(os.getpid(), signal.SIGKILL)



import asyncio
import concurrent.futures
import fcntl
import hashlib
import json
import multiprocessing
import os
import threading
from pathlib import Path
from typing import Any, Dict, Optional, cast

from fastapi import HTTPException
from outlines.fsm.guide import RegexGuide
from outlines.models.transformers import TransformerTokenizer
from outlines.processors.structured import JSONLogitsProcessor
from referencing.exceptions import Unresolvable
from transformers import PreTrainedTokenizerFast

from briton.proto import StatesToTokens, TokenToNextState


# This must be defined globally to be used in the forked processes
outlines_tokenizer = None


def worker(
    vocab_size: int, end_id: int, tools_id: Optional[int], schema: Dict[str, Any], output_path: Path
):
    """
    Worker to create FSM and serialize to protobuf.

    Args:
        vocab_size (int): The size of the vocabulary.
        end_id (int): The end-of-sequence token ID.
        schema (Dict[str, Any]): The schema used by the JSONLogitsProcessor.
        output_path (Path): The path where the serialized protocol buffer will be written.
    """
    assert outlines_tokenizer is not None
    try:
        logits_processor = JSONLogitsProcessor(schema, outlines_tokenizer)
    except (Unresolvable, ValueError, TypeError) as e:
        raise NotImplementedError(str(e))
    guide = cast(RegexGuide, logits_processor.fsm)
    states_to_tokens = {}
    for state, token_to_next_state in guide.states_to_token_maps.items():
        states_to_tokens[state] = TokenToNextState(token_to_next_state=token_to_next_state)
    states_to_tokens_pb = StatesToTokens(
        states_to_tokens=states_to_tokens,
        vocab_size=vocab_size,
        eos_token_id=end_id,
        tools_id=tools_id,
    )
    if not output_path.exists():
        try:
            # Open the file with flags to protect against concurrent writes.
            # O_CREAT: Create the file if it does not exist.
            # O_EXCL: Ensure that this call creates the file exclusively. If the file already exists, the call will fail.
            # O_WRONLY: Open the file for write-only access.
            fd = os.open(output_path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
            with os.fdopen(fd, "wb") as f:
                fcntl.flock(f, fcntl.LOCK_EX)
                f.write(states_to_tokens_pb.SerializeToString())
                fcntl.flock(f, fcntl.LOCK_UN)
        except FileExistsError:
            pass


# Dummy task to iniitialize processses in the ProcessPoolExecutor used for FSM generation
def dummy_task():
    pass


class FsmCache:
    def __init__(
        self,
        cache_dir: Path,
        tokenizer: PreTrainedTokenizerFast,
        max_workers: int,
        tools_id: Optional[int],
    ):
        self._cache_dir = cache_dir
        if not self._cache_dir.exists():
            self._cache_dir.mkdir(parents=True, exist_ok=True)
        self._cache = set(f.name for f in self._cache_dir.iterdir() if f.is_file())
        self._lock = threading.Lock()
        self._tokenizer = tokenizer
        self._vocab_size = len(getattr(self._tokenizer, "vocab"))
        self._eos_token_id = getattr(self._tokenizer, "eos_token_id")
        self._tools_id = tools_id

        # Concurrent FSM generation initialization
        # Make sure we fork because (1) it's faster and (2) it seems that spawning
        # ends up being sequential
        multiprocessing.set_start_method("fork", force=True)
        global outlines_tokenizer
        outlines_tokenizer = TransformerTokenizer(tokenizer)  # type: ignore
        # This is very important. The first time JSONLogitsProcessor is called, some library-wide
        # initializations are done in memory (that take 5s). By doing it before we fork, we avoid paying
        # that cost for each forked process.
        _ = JSONLogitsProcessor({"properties": {}}, outlines_tokenizer)
        self._executor = concurrent.futures.ProcessPoolExecutor(max_workers=max_workers)
        # We must create all processes BEFORE the GRPC python client is started to avoid errors
        # forking from the process GRPC is running in
        for _ in range(max_workers):
            self._executor.submit(dummy_task)

    @property
    def cache_dir(self) -> str:
        return str(self._cache_dir)

    async def add_schema(self, schema: Dict[str, Any]) -> str:
        schema_str = json.dumps(schema)
        schema_hash = hashlib.sha256(schema_str.encode()).hexdigest()
        if schema_hash not in self._cache:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(
                self._executor,
                worker,
                self._vocab_size,
                self._eos_token_id,
                self._tools_id,
                schema,
                self._cache_dir / schema_hash,
            )
            with self._lock:
                self._cache.add(schema_hash)
        return schema_hash


async def add_schema_to_cache(fsm_cache: FsmCache, schema: Dict[str, Any]) -> str:
    try:
        schema_hash = await fsm_cache.add_schema(schema)
    except NotImplementedError as ex:
        raise HTTPException(status_code=400, detail=str(ex))
    return schema_hash



import asyncio
import functools
import logging
import time
from typing import Awaitable, Callable, Coroutine, List, Optional

from starlette.requests import Request as StarletteRequest

from briton.async_util import (
    get_all_items,
    predicate_with_timeout,
    retry_predicate,
    set_event_loop_if_not_exist,
)


logger = logging.getLogger(__name__)


ProcessorFunctionType = Callable[
    [List["SpecDecRequest"]], Coroutine[None, None, List["SpecDecRequest"]]
]


class SpecDecRequest:
    id: int
    input_ids: List[int]
    draft_ids: List[int]
    model_input: dict
    produced_token_count: int
    accepted_token_count: int
    target_time_ms: int
    draft_time_ms: int
    draft_token_count: int
    start_time: float
    iterations: int
    is_cancelled_fn: Callable[[], Awaitable[bool]]
    output_schema_hash: Optional[str]
    fsm_state: Optional[int]
    force_tools: Optional[bool]

    prompt_tokens: int

    def __init__(
        self,
        id: int,
        input_ids: List[int],
        draft_ids: List[int],
        model_input: dict,
        is_cancelled_fn: Callable[[], Awaitable[bool]],
        output_schema_hash: Optional[str] = None,
        force_tools: Optional[bool] = None,
    ):
        self.id = id
        self.input_ids = input_ids
        self.draft_ids = draft_ids
        self.model_input = model_input
        self.queue = asyncio.Queue()
        self.produced_token_count = 0
        self.accepted_token_count = 0
        self.target_time_ms = 0
        self.draft_time_ms = 0
        self.draft_token_count = 0
        self.start_time = time.time()
        self.iterations = 0
        self.is_cancelled_fn = is_cancelled_fn
        self.output_schema_hash = output_schema_hash
        self.fsm_state = None
        self.force_tools = force_tools

        # The number of tokens in the prompt; only updated at initialization.
        self.prompt_tokens = len(self.input_ids)

    def completion_tokens(self) -> int:
        return self.produced_token_count


def handle_request_exception(func):
    """Informs request async queue of the exception and returns None.

    Important to return None so it doesn't terminate the worker loop.
    """

    @functools.wraps(func)
    async def wrapper(req, *args, **kwargs):
        try:
            return await func(req, *args, **kwargs)
        except Exception as e:
            print(f"Unexpected error: {e}")
            req.queue.put_nowait(e)
            return None

    return wrapper


async def worker(
    target_queue: asyncio.Queue,
    target_processor: ProcessorFunctionType,
    draft_queue: asyncio.Queue,
    draft_processor: ProcessorFunctionType,
    stop_event: asyncio.Event,
):
    """Main loop in rail design.

    We alternate between target and draft model, reading and processing items
    from the respective queues.
    """
    while not stop_event.is_set():
        # Target
        # Wait for input
        target_item_get_task = asyncio.create_task(target_queue.get())
        stop_event_wait_task = asyncio.create_task(stop_event.wait())
        done, _ = await asyncio.wait(
            [target_item_get_task, stop_event_wait_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        if stop_event.is_set():
            print("Stopping worker")
            target_item_get_task.cancel()
            break

        for task in done:
            target_item = task.result()
            remaining_target_items = get_all_items(target_queue)
            items = [target_item] + remaining_target_items

        target_cont_items = await target_processor(items)
        for item in target_cont_items:
            draft_queue.put_nowait(item)

        # Draft
        items = get_all_items(draft_queue)
        draft_cont_items = await draft_processor(items)
        for item in draft_cont_items:
            target_queue.put_nowait(item)


def pick_tokens_from_end(num_input_ids, num_draft_max_input_tokens: int) -> int:
    """
    Draft model may have a smaller window then target. In that case we clamp to
    the shorter draft window. We do this so that the start point is stable, and
    only moves back when we go over the max.

    Net effect of this should be that we pick all tokens until just before max.
    Afterwards we start skipping first `max draft tokens // 2` tokens. And later
    more and more multiples of `max draft tokens // 2` initial tokens.

    Idea is to keep the tokens passed to draft stable, to benefit from kv cache.
    We keep at least

    Returns starting point to pick input_ids from.
    """
    half_max = num_draft_max_input_tokens // 2
    half_maxes_to_skip = max(0, ((num_input_ids - 1) // half_max) - 1)
    return half_max * half_maxes_to_skip


def monitor_thread_fn(
    monitor_fn: callable,
    on_fail: callable,
    should_continue: Optional[Callable] = None,
    period_secs: int = 60,
    max_retries: int = 3,
    timeout_secs: int = 30,
    retry_delay_secs: int = 2,
):
    set_event_loop_if_not_exist()
    if should_continue is None:
        should_continue = lambda: True
    logger.info("Starting monitor thread")
    asyncio.run(
        monitor(
            monitor_fn=monitor_fn,
            on_fail=on_fail,
            should_continue=should_continue,
            period_secs=period_secs,
            max_retries=max_retries,
            timeout_secs=timeout_secs,
            retry_delay_secs=retry_delay_secs,
        )
    )


async def monitor(
    monitor_fn: Callable[[], Awaitable[bool]],
    on_fail: Callable[[], None],
    should_continue: Optional[Callable[[], bool]],
    period_secs: int = 60,
    max_retries: int = 3,
    timeout_secs: int = 60,
    retry_delay_secs: int = 2,
):
    """Monitor by calling a function.
    Args:
        monitor_fn: The function to call to check.
        on_fail: Function to call when the model is unresponsive.
        should_continue: Function to determine if monitoring should continue.
        period_secs: How often to check the model.
        retries: How many times to retry.
    """
    while should_continue():
        if not await _check_with_timeout_and_retry(
            monitor_fn,
            timeout_secs=timeout_secs,
            max_retries=max_retries,
            retry_delay_secs=retry_delay_secs,
        ):
            on_fail()
            break
        else:
            await asyncio.sleep(period_secs)


async def _check_with_timeout_and_retry(
    monitor_fn: Callable[[], Awaitable[bool]],
    timeout_secs: int = 60,
    max_retries: int = 3,
    retry_delay_secs: int = 2,
) -> Awaitable[bool]:
    with_timeout = predicate_with_timeout(monitor_fn, timeout_secs=timeout_secs)
    with_retry = retry_predicate(with_timeout, max_retries=max_retries, delay_secs=retry_delay_secs)
    return await with_retry()


async def test_predict(async_predict_fn: Callable) -> bool:
    try:
        model_input = {
            "prompt": "test",
            "max_tokens": 1,
        }
        fastapi_request = StarletteRequest(
            scope={
                "type": "http",
                "method": "POST",
            }
        )
        fastapi_request.is_disconnected = lambda: False
        await async_predict_fn(model_input, fastapi_request)
        return True
    except Exception as e:
        logger.info(f"Error during prediction test: {e}")
        return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: briton.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x62riton.proto\x12\x06\x62riton\"r\n\x06Tensor\x12#\n\x05shape\x18\x01 \x01(\x0b\x32\x14.briton.Tensor.Shape\x12\x1f\n\x05\x64type\x18\x02 \x01(\x0e\x32\x10.briton.DataType\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x1a\x14\n\x05Shape\x12\x0b\n\x03\x64im\x18\x01 \x03(\x03\"\x1c\n\x05\x42\x61tch\x12\x13\n\x0brequest_ids\x18\x01 \x03(\x03\"\x98\x08\n\x10InferenceRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\x03\x12\x17\n\ninput_text\x18\x02 \x01(\tH\x00\x88\x01\x01\x12\x11\n\tinput_ids\x18\x03 \x03(\x05\x12\x1f\n\x12request_output_len\x18\x05 \x01(\rH\x01\x88\x01\x01\x12\x13\n\x06\x65nd_id\x18\x06 \x01(\rH\x02\x88\x01\x01\x12\x13\n\x06pad_id\x18\x07 \x01(\rH\x03\x88\x01\x01\x12\x17\n\nbeam_width\x18\n \x01(\rH\x04\x88\x01\x01\x12\x18\n\x0btemperature\x18\x0b \x01(\x02H\x05\x88\x01\x01\x12\x1a\n\rruntime_top_k\x18\x0c \x01(\rH\x06\x88\x01\x01\x12\x1a\n\rruntime_top_p\x18\r \x01(\x02H\x07\x88\x01\x01\x12\x18\n\x0blen_penalty\x18\x0e \x01(\x02H\x08\x88\x01\x01\x12\x1f\n\x12repetition_penalty\x18\x0f \x01(\x02H\t\x88\x01\x01\x12\x1d\n\x10presence_penalty\x18\x10 \x01(\x02H\n\x88\x01\x01\x12\x11\n\tbad_words\x18\x11 \x03(\t\x12\x12\n\nstop_words\x18\x12 \x03(\t\x12\x19\n\x0clora_task_id\x18\x13 \x01(\x04H\x0b\x88\x01\x01\x12)\n\x0clora_weights\x18\x14 \x01(\x0b\x32\x0e.briton.TensorH\x0c\x88\x01\x01\x12(\n\x0blora_config\x18\x15 \x01(\x0b\x32\x0e.briton.TensorH\r\x88\x01\x01\x12\x18\n\x0brandom_seed\x18\x16 \x01(\x03H\x0e\x88\x01\x01\x12\x17\n\x0f\x64raft_input_ids\x18\x17 \x03(\x05\x12\x1f\n\x12output_schema_hash\x18\x18 \x01(\tH\x0f\x88\x01\x01\x12\x16\n\tfsm_state\x18\x19 \x01(\x05H\x10\x88\x01\x01\x12\x18\n\x0b\x66orce_tools\x18\x1a \x01(\x08H\x11\x88\x01\x01\x12 \n\x13write_output_states\x18\x1b \x01(\x08H\x12\x88\x01\x01\x12!\n\x05\x62\x61tch\x18\x1c \x01(\x0b\x32\r.briton.BatchH\x13\x88\x01\x01\x42\r\n\x0b_input_textB\x15\n\x13_request_output_lenB\t\n\x07_end_idB\t\n\x07_pad_idB\r\n\x0b_beam_widthB\x0e\n\x0c_temperatureB\x10\n\x0e_runtime_top_kB\x10\n\x0e_runtime_top_pB\x0e\n\x0c_len_penaltyB\x15\n\x13_repetition_penaltyB\x13\n\x11_presence_penaltyB\x0f\n\r_lora_task_idB\x0f\n\r_lora_weightsB\x0e\n\x0c_lora_configB\x0e\n\x0c_random_seedB\x15\n\x13_output_schema_hashB\x0c\n\n_fsm_stateB\x0e\n\x0c_force_toolsB\x16\n\x14_write_output_statesB\x08\n\x06_batch\"i\n\x13InferenceAnswerPart\x12\x12\n\nrequest_id\x18\x01 \x01(\x03\x12\x13\n\x0boutput_text\x18\x02 \x01(\t\x12\x12\n\noutput_ids\x18\x03 \x03(\x05\x12\x15\n\routput_states\x18\x04 \x03(\x05\"\xe2\x01\n\nAddedToken\x12\x14\n\x07\x63ontent\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x18\n\x0bsingle_word\x18\x02 \x01(\x08H\x01\x88\x01\x01\x12\x13\n\x06lstrip\x18\x03 \x01(\x08H\x02\x88\x01\x01\x12\x13\n\x06rstrip\x18\x04 \x01(\x08H\x03\x88\x01\x01\x12\x17\n\nnormalized\x18\x05 \x01(\x08H\x04\x88\x01\x01\x12\x14\n\x07special\x18\x06 \x01(\x08H\x05\x88\x01\x01\x42\n\n\x08_contentB\x0e\n\x0c_single_wordB\t\n\x07_lstripB\t\n\x07_rstripB\r\n\x0b_normalizedB\n\n\x08_special\"1\n\x0b\x41\x64\x64\x65\x64Tokens\x12\"\n\x06tokens\x18\x01 \x03(\x0b\x32\x12.briton.AddedToken\"\xe9\t\n\x0c\x42ritonConfig\x12\x13\n\x0b\x65ngine_path\x18\x01 \x01(\t\x12\x14\n\x0chf_tokenizer\x18\x02 \x01(\t\x12\x11\n\x04port\x18\x03 \x01(\rH\x00\x88\x01\x01\x12N\n\x16\x62\x61tch_scheduler_policy\x18\x05 \x01(\x0e\x32).briton.BritonConfig.BatchSchedulerPolicyH\x01\x88\x01\x01\x12\x1f\n\x12\x65nable_trt_overlap\x18\x06 \x01(\x08H\x02\x88\x01\x01\x12)\n\x1cmax_tokens_in_paged_kv_cache\x18\n \x01(\x04H\x03\x88\x01\x01\x12+\n\x1ekv_cache_free_gpu_mem_fraction\x18\x0b \x01(\x02H\x04\x88\x01\x01\x12!\n\x14medusa_decoding_mode\x18\x0c \x01(\x08H\x05\x88\x01\x01\x12#\n\x16\x65nable_chunked_context\x18\r \x01(\x08H\x06\x88\x01\x01\x12\"\n\x15\x65nable_kv_cache_reuse\x18\x0e \x01(\x08H\x07\x88\x01\x01\x12\'\n\x1akv_cache_host_memory_bytes\x18\x0f \x01(\x04H\x08\x88\x01\x01\x12(\n\x1blora_cache_max_adapter_size\x18\x10 \x01(\x04H\t\x88\x01\x01\x12,\n\x1flora_cache_optimal_adapter_size\x18\x11 \x01(\x04H\n\x88\x01\x01\x12+\n\x1elora_cache_gpu_memory_fraction\x18\x12 \x01(\x02H\x0b\x88\x01\x01\x12)\n\x1clora_cache_host_memory_bytes\x18\x13 \x01(\x04H\x0c\x88\x01\x01\x12\x1a\n\rfsm_cache_dir\x18\x14 \x01(\tH\r\x88\x01\x01\x12\x1e\n\x11max_batch_wait_ms\x18\x15 \x01(\rH\x0e\x88\x01\x01\x12.\n\x0c\x61\x64\x64\x65\x64_tokens\x18\x16 \x01(\x0b\x32\x13.briton.AddedTokensH\x0f\x88\x01\x01\x12\x1b\n\x0emax_num_tokens\x18\x17 \x01(\rH\x10\x88\x01\x01\"D\n\x14\x42\x61tchSchedulerPolicy\x12\x13\n\x0fMAX_UTILIZATION\x10\x00\x12\x17\n\x13GUARANTEED_NO_EVICT\x10\x01\x42\x07\n\x05_portB\x19\n\x17_batch_scheduler_policyB\x15\n\x13_enable_trt_overlapB\x1f\n\x1d_max_tokens_in_paged_kv_cacheB!\n\x1f_kv_cache_free_gpu_mem_fractionB\x17\n\x15_medusa_decoding_modeB\x19\n\x17_enable_chunked_contextB\x18\n\x16_enable_kv_cache_reuseB\x1d\n\x1b_kv_cache_host_memory_bytesB\x1e\n\x1c_lora_cache_max_adapter_sizeB\"\n _lora_cache_optimal_adapter_sizeB!\n\x1f_lora_cache_gpu_memory_fractionB\x1f\n\x1d_lora_cache_host_memory_bytesB\x10\n\x0e_fsm_cache_dirB\x14\n\x12_max_batch_wait_msB\x0f\n\r_added_tokensB\x11\n\x0f_max_num_tokens\"\x98\x01\n\x10TokenToNextState\x12K\n\x13token_to_next_state\x18\x01 \x03(\x0b\x32..briton.TokenToNextState.TokenToNextStateEntry\x1a\x37\n\x15TokenToNextStateEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"\xf5\x01\n\x0eStatesToTokens\x12\x44\n\x10states_to_tokens\x18\x01 \x03(\x0b\x32*.briton.StatesToTokens.StatesToTokensEntry\x12\x12\n\nvocab_size\x18\x02 \x01(\x05\x12\x14\n\x0c\x65os_token_id\x18\x03 \x01(\x05\x12\x15\n\x08tools_id\x18\x04 \x01(\x05H\x00\x88\x01\x01\x1aO\n\x13StatesToTokensEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.briton.TokenToNextState:\x02\x38\x01\x42\x0b\n\t_tools_id*\xa8\x01\n\x08\x44\x61taType\x12\x0e\n\nDT_INVALID\x10\x00\x12\x0b\n\x07\x44T_INT4\x10\x01\x12\x0b\n\x07\x44T_INT8\x10\x02\x12\x0c\n\x08\x44T_UINT8\x10\x03\x12\x0c\n\x08\x44T_INT32\x10\x04\x12\x0c\n\x08\x44T_INT64\x10\x05\x12\x0e\n\nDT_FLOAT16\x10\n\x12\x0f\n\x0b\x44T_BFLOAT16\x10\x0b\x12\x0e\n\nDT_FLOAT32\x10\x0c\x12\n\n\x06\x44T_FP8\x10\r\x12\x0b\n\x07\x44T_BOOL\x10\x14\x32L\n\x06\x42riton\x12\x42\n\x05Infer\x12\x18.briton.InferenceRequest\x1a\x1b.briton.InferenceAnswerPart\"\x00\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'briton_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY']._loaded_options = None
  _globals['_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY']._serialized_options = b'8\001'
  _globals['_STATESTOTOKENS_STATESTOTOKENSENTRY']._loaded_options = None
  _globals['_STATESTOTOKENS_STATESTOTOKENSENTRY']._serialized_options = b'8\001'
  _globals['_DATATYPE']._serialized_start=3272
  _globals['_DATATYPE']._serialized_end=3440
  _globals['_TENSOR']._serialized_start=24
  _globals['_TENSOR']._serialized_end=138
  _globals['_TENSOR_SHAPE']._serialized_start=118
  _globals['_TENSOR_SHAPE']._serialized_end=138
  _globals['_BATCH']._serialized_start=140
  _globals['_BATCH']._serialized_end=168
  _globals['_INFERENCEREQUEST']._serialized_start=171
  _globals['_INFERENCEREQUEST']._serialized_end=1219
  _globals['_INFERENCEANSWERPART']._serialized_start=1221
  _globals['_INFERENCEANSWERPART']._serialized_end=1326
  _globals['_ADDEDTOKEN']._serialized_start=1329
  _globals['_ADDEDTOKEN']._serialized_end=1555
  _globals['_ADDEDTOKENS']._serialized_start=1557
  _globals['_ADDEDTOKENS']._serialized_end=1606
  _globals['_BRITONCONFIG']._serialized_start=1609
  _globals['_BRITONCONFIG']._serialized_end=2866
  _globals['_BRITONCONFIG_BATCHSCHEDULERPOLICY']._serialized_start=2350
  _globals['_BRITONCONFIG_BATCHSCHEDULERPOLICY']._serialized_end=2418
  _globals['_TOKENTONEXTSTATE']._serialized_start=2869
  _globals['_TOKENTONEXTSTATE']._serialized_end=3021
  _globals['_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY']._serialized_start=2966
  _globals['_TOKENTONEXTSTATE_TOKENTONEXTSTATEENTRY']._serialized_end=3021
  _globals['_STATESTOTOKENS']._serialized_start=3024
  _globals['_STATESTOTOKENS']._serialized_end=3269
  _globals['_STATESTOTOKENS_STATESTOTOKENSENTRY']._serialized_start=3177
  _globals['_STATESTOTOKENS_STATESTOTOKENSENTRY']._serialized_end=3256
  _globals['_BRITON']._serialized_start=3442
  _globals['_BRITON']._serialized_end=3518
# @@protoc_insertion_point(module_scope)
